//coded by Feifei Li, contact: lifeifei@cs.bu.edu


#include "doublee.h"
#include <stdio.h>
#include <stdlib.h>
#include <iostream>

ostream & operator<<( ostream &os, Doublee &db ) 
{ 
        os << db.a<< "e" <<db.b; 
        return ( os ); 
}

Doublee::Doublee()
{                 
       this->a=0.0;
       this->b=0.0;
}

Doublee::Doublee(double a, double b)
{
       this->a=a;
       this->b=b;
}

Doublee::Doublee(const Doublee& db)
{
       this->a=db.a;
       this->b=db.b;                          
}

Doublee::Doublee(double doubleval)
{
       b=ceil(log10(doubleval));       
       a=doubleval/pow(10.0,b);
}

Doublee::~Doublee()
{}

double Doublee::getDouble()
{
       return a*pow(10.0, b);       
}

void Doublee::adjustExp(Doublee& t)
{
      if (this->b!=t.b)
      {
          if (this->b<t.b)
          {
             t.a=t.a*pow(10.0, (t.b-this->b));                
             t.b=this->b;
          }
          else
          {
             t.a=t.a/pow(10.0,(this->b-t.b));                
             t.b=this->b;
          }
      }
}

void Doublee::adjustAB()
{
      double tmp=log10(this->a);
      double scale = ceil(tmp);
      this->a=this->a/pow(10,scale);
      this->b=this->b+scale;
}

void Doublee::add(Doublee t)
{
      this->adjustExp(t);
      this->a=this->a+t.a;
      if (this->a<DB_MIN || this->a>DB_MAX)
         this->adjustAB();
}

void Doublee::minus(Doublee t)
{
     this->adjustExp(t);
     this->a=this->a-t.a;
     if (this->a<DB_MIN || this->a>DB_MAX)
         this->adjustAB();
}

void Doublee::multiply(Doublee t)
{
     this->a=this->a*t.a;
     this->b=this->b+t.b;
     if (this->a<DB_MIN || this->a>DB_MAX)
         this->adjustAB();
}

void Doublee::divide(Doublee t)
{
     this->a=this->a/t.a;
     this->b=this->b-t.b;
     if (this->a<DB_MIN || this->a>DB_MAX)
         this->adjustAB();
}

void Doublee::assign(Doublee t)
{
     this->a=t.a;
     this->b=t.b;
}

Doublee Doublee::operator=(double value)
{
       b=ceil(log10(value));       
       a=value/pow(10.0,b);
       return (*this);
}

Doublee Doublee::operator+(double value)
{
      Doublee t(value);
      Doublee s(*this);
      s.add(t);
      return s;
}

Doublee Doublee::operator-(double value)
{
       Doublee t(value);
       Doublee s(*this);
       s.minus(t);
       return s;
}

Doublee Doublee::operator/(double value)
{
       Doublee t(value);
       Doublee s(*this);
       s.divide(t);
       return s;
}

Doublee Doublee::operator*(double value)
{
      Doublee t(value);
      Doublee s(*this);
      s.multiply(t);
      return s;
}

bool Doublee::operator>(double value)
{
     Doublee t(value);
     this->adjustExp(t);
     return (this->a>t.a);
}

bool Doublee::operator<(double value)
{
     Doublee t(value);
     this->adjustExp(t);
     return (this->a<t.a);
}

bool Doublee::operator>=(double value)
{
     Doublee t(value);
     this->adjustExp(t);
     return (this->a>=t.a);
}

bool Doublee::operator<=(double value)
{
     Doublee t(value);
     this->adjustExp(t);
     return (this->a<=t.a);
}

bool Doublee::operator==(double value)
{
     Doublee t(value);
     this->adjustExp(t);
     return (this->a==t.a);
}

bool Doublee::operator!=(double value)
{
     Doublee t(value);
     this->adjustExp(t);
     return (this->a!=t.a);
}

bool Doublee::operator<(Doublee value)
{
     this->adjustExp(value);
     return (this->a<value.a);
}

bool Doublee::operator>(Doublee value)
{
     this->adjustExp(value);
     return (this->a>value.a);
}

bool Doublee::operator<=(Doublee value)
{
     this->adjustExp(value);
     return (this->a<=value.a);
}

bool Doublee::operator>=(Doublee value)
{
     this->adjustExp(value);
     return (this->a>=value.a);
}

bool Doublee::operator==(Doublee value)
{
     this->adjustExp(value);
     return (this->a==value.a);
}

bool Doublee::operator!=(Doublee value)
{
     this->adjustExp(value);
     return (this->a!=value.a);
}
    
Doublee Doublee::operator=(Doublee value)
{
    this->assign(value);
    return (*this);
}

Doublee Doublee::operator+(Doublee value)
{
      Doublee s(*this);
      s.add(value);
      return s;
}

Doublee Doublee::operator-(Doublee value)
{
      Doublee s(*this);
      s.minus(value);
      return s;
}

Doublee Doublee::operator/(Doublee value)
{
     Doublee s(*this);
     s.divide(value);
     return s;
}

Doublee Doublee::operator*(Doublee value)
{
     Doublee s(*this);
     s.multiply(value);
     return s;
}


    
Doublee operator+(double val, Doublee db)
{
        Doublee t(val);
        return (t+db);
}

Doublee operator-(double val, Doublee db)
{
        Doublee t(val);
        return (t-db);
}

Doublee operator*(double val, Doublee db)
{
        Doublee t(val);
        return (t*db);
}

Doublee operator/(double val, Doublee db)
{
        Doublee t(val);
        return t/db;
}

bool operator>(double val, Doublee db)
{
        Doublee t(val);
        db.adjustExp(t);
        return (db.a>t.a);
}

bool operator<(double val, Doublee db)
{
        Doublee t(val);
        db.adjustExp(t);
        return (db.a<t.a);
}

bool operator>=(double val, Doublee db)
{
        Doublee t(val);
        db.adjustExp(t);
        return (db.a>=t.a);
}

bool operator<=(double val, Doublee db)
{
        Doublee t(val);
        db.adjustExp(t);
        return (db.a<=t.a);
}

bool operator==(double val, Doublee db)
{
        Doublee t(val);
        db.adjustExp(t);
        return (db.a==t.a);
}

bool operator!=(double val, Doublee db)
{
        Doublee t(val);
        db.adjustExp(t);
        return (db.a!=t.a);
}


