//coded by Feifei Li, contact: lifeifei@cs.bu.edu

#include "ukrankindep.h"

UkrankIndep::UkrankIndep(int k)
{
   this->k=k;    
   this->db=NULL;
   elapsed=0.0;                       
}

UkrankIndep::~UkrankIndep()
{
   if (db!=NULL)
      delete db;                                                   
}

bool UkrankIndep::initDB(unsigned int N, char* input_file)
{
   db= new Udb(N, input_file);
   if (!(db->fin))
      return false;
   db->readRecords();
   //db->sortByScore();
   return true;     
}

void UkrankIndep::findTopk()
{
   unsigned int i=0;
   unsigned int index;
   unsigned int j=0;
   bool stop=false;
   Rij.resize(this->k+1);
   topk.resize(this->k+1);
   krank.resize(this->k+1);
   for (i=0;i<(unsigned int)k+1;i++)
   {
       Rij[i]=0.0;
       krank[i]=0.0;
   }    
   Rij[0]=1.0;
   i=1;
   t1 = times(&st);
   gettimeofday (&start,NULL);
   while (true)
   {
          Tuple tp=db->tuples[i-1];
          index=k;
          if (i<(unsigned int)k)
              index=i;
          for (j=1;j<index+1;j++)
          {
              double newconf=tp.confidence*Rij[j-1];   
              if (krank[j]<newconf)
              {
                  krank[j]=newconf;
                  topk[j]=tp.sortid;
              }      
          }
          for (j=index;j>=1;j--)
              Rij[j]=tp.confidence*Rij[j-1]+(1-tp.confidence)*Rij[j];
          Rij[0]=(1-tp.confidence)*Rij[0];
          if (i>=(unsigned int)k)
          {
              stop=checkTerminate(krank, Rij);
             if (stop)
             {
                 this->n=i;
                  break;
              }
          }  
          i++;
          if (i>db->N)
          {
            this->n=db->N;
            break;
          }
   }
   gettimeofday (&end,NULL);
   t2 = times(&ed);
   usertime=((double)(ed.tms_utime-st.tms_utime))/(HZ);
   systemtime=((double)(ed.tms_stime-st.tms_stime))/(HZ);
   realtime=((double)(t2-t1))/(HZ);
   db->timeval_subtract(&tdiff, &end, &start);
   elapsed=tdiff.tv_sec+tdiff.tv_usec/1000000.0;
   getMemsize();
}

void UkrankIndep::printStatistics()
{	
    
    cout<<endl<<"scan depth: "<<this->n<<endl<<endl;
    printf("user time=%f \n", usertime);
    printf("system time=%f \n", systemtime);
    printf("real time=%f \n", realtime);
    printf("clock time=%f \n", elapsed);
    cout<<"max mem usage(in bytes)= "<<memusage<<endl;
    cout<<"max mem usage(in KB)= "<<(double)memusage/1024.0<<endl;
    cout<<"max mem usage(in MB)= "<<(double)memusage/(1024.0*1024.0)<<endl; 
}

bool UkrankIndep::checkTerminate(vector<double> krank, vector<double>Rij)
{
     int i=0;
     bool stop=true;
     double maxrnl=0.0;
     for (i=1;i<=this->k;i++)
     {
         if (maxrnl<Rij[i-1])
              maxrnl=Rij[i-1];
         if (krank[i]<maxrnl)
         {
              stop=false;
              break;                      
          }  
     }
     return stop;
}

void UkrankIndep::printTopk()
{
     int i=0;
     for (i=1;i<=k;i++)
     {
         Tuple tp=db->tuples[topk[i]];
         cout<<tp<<" "<<krank[i]<<endl;
     } 
}

void UkrankIndep::printRij()
{
     int i=0;
     for (i=0;i<k+1;i++)
         cout<<Rij[i]<<endl;
}


void UkrankIndep::getMemsize()
{
     //it's bounded by Rij and top k tuples..
     memusage=k*Tuple::size(false); 
     memusage+=k*8;
}
