//coded by Feifei Li, contact: lifeifei@cs.bu.edu

#include "utopkcor.h"

bool compare_tuple_score1(Tuple t1, Tuple t2)
{
    return t1.sortid<t2.sortid;     
} 


UtopkCor::UtopkCor(int k)
{
   this->k=k;    
   this->db=NULL;
   elapsed=0.0;                       
}

UtopkCor::~UtopkCor()
{
   if (db!=NULL)                      
      delete db;  
   topk.clear();
   __gnu_cxx::hash_map<unsigned int, Tuple*, GnuHash>::iterator it=hatDi.begin();
   while (it!=hatDi.end())
   {
         delete (*it).second;
         it++;
   }
   hatDi.clear();     
   
}

bool UtopkCor::initDB(unsigned int N, char* input_file, char* xtuple_file, int xdegree, unsigned int numxtuple)
{
   db= new Udb(N, input_file);
   if (!(db->fin))
      return false;
   db->readRecords();
   db->buildXtuples(xtuple_file, xdegree, numxtuple);
   //db->sortByScore(); not necessary if the input db has already sorted, which is the case for our generator.
   return true;     
}

void UtopkCor::findTopk()
{
   unsigned int i=0;
   long double terminator=1.0;
   long double largest_rhoi=0.0;
   long double rhoi=1.0;
   unsigned int xid=0;
   bool newdominator=false;
   bool firstk=true;
   //bool dom_in_topk=false;
   __gnu_cxx::hash_map<unsigned int, Tuple*, GnuHash>::iterator it;	        
   map<Tuple*, Tuple*, PrioritizeTask>::iterator it_topk;	        
   
   t1 = times(&st);
   gettimeofday(&start,NULL);
   while (true)
   {
         Tuple tp=db->tuples[i];
         newdominator=false;
         Tuple* dom=NULL;
         
         if (db->xtuples[tp.sortid].size()!=0)
         {
                   xid=db->xtuples[tp.sortid][0];
                   if (xid<=i)
                   {
                      it=hatDi.find(db->tuples[xid].id);
                      if (it==hatDi.end())
                      {
                         cout<<"sth is wrong..."<<endl;
                         break;
                      }   
                      dom=(*it).second;
                      if (dom->confidence<tp.confidence)
                      {                        
                           it_topk=topk.find(dom);
                           if (it_topk!=topk.end())
                           {
                              //cout<<"found dom and get rid of it "<<endl;
                              topk.erase(it_topk); //erase this dominator from the topk (binary tree) as its value gets updated   
                              rhoi=rhoi/dom->confidence;
                           }
                           else
                              rhoi=rhoi/dom->qitj;
                           tp.qitj=dom->qitj-tp.confidence; 
                           if (fabs(tp.qitj-0.0)<ZERO)                               
                                 tp.qitj=0.0;  
                           dom->copy(tp);
                           newdominator=false;
                          // tp is the new dominator for all its siblings (that has appeared bfe i) in the same xtuple                                                 
                      }
                      else
                      {
                            newdominator=false;
                            it_topk=topk.find(dom);
                            if (it_topk!=topk.end())
                            {
                                topk.erase(it_topk);
                                rhoi=rhoi/dom->confidence;
                            }
                            else
                                rhoi=rhoi/dom->qitj;
                            

                            if (dom->confidence>dom->qitj)
                                 terminator=terminator/dom->confidence;
                            else
                                 terminator=terminator/dom->qitj;
                            dom->qitj=dom->qitj-tp.confidence;
                            if (fabs(dom->qitj-0.0)<ZERO)
                               dom->qitj=0.0;
                      }   
                   }
                   else
                   {
                           //cout<<"xtuple without preceedings... "<<tp<<endl;
                           tp.qitj=1-tp.confidence;
                           newdominator=true;
                        //handle the case where this tp has xtuple, but all its siblings have not appeared yet. 
                       //He is the fisrt guy, hence automatically becomes the dominator for this xtuple.
                   }
         }
         else
         {
             //cout<<"non xtuple "<<tp<<endl;
             tp.qitj=1-tp.confidence;
             newdominator=true;
             // it's not a xtuple, hence it automatically becomes its dominator.
         }  
         //Di.push_back(tp);
         if (newdominator)
         {                        
               dom=new Tuple(tp);
               hatDi.insert(pair<unsigned int, Tuple*>(dom->id, dom));
               //cout<<"new dom..."<<dom->sortid<<endl; 
         }
         
         if (dom->confidence>dom->qitj)
            terminator=terminator*dom->confidence;
         else
            terminator=terminator*dom->qitj; 
         
         
         topk.insert(pair<Tuple*, Tuple*>(dom, dom));
         
         if (!firstk)
         {
               if (topk.size()>(unsigned int)k)
               {
                   it_topk=topk.end();
                   it_topk--;
                   Tuple* poped=(*it_topk).second;
                   topk.erase(it_topk);
                   rhoi=(rhoi/poped->confidence)*dom->confidence*poped->qitj;
                   if (poped->sortid!=dom->sortid)
                       update_snapshot_topk.push_back(dom);
               }
               else
                   rhoi=rhoi*dom->confidence;
               if (rhoi>largest_rhoi)
               {    
                   largest_rhoi=rhoi;
                   //snapshot_topk=topk;     
                   if (update_snapshot_topk.size()>=(unsigned int)k)
                   {
                         update_snapshot_topk.clear();
                         snapshot_topk=topk;     
                   }
                   largest_rhoi_index=dom->sortid;
               }
         }
         else
         {
                rhoi=rhoi*dom->confidence;
                if (topk.size()==(unsigned int)k)
                {
                   largest_rhoi=rhoi;
                   snapshot_topk=topk;
                   firstk=false;
                }
         } 
         
         //debug information
         /*cout<<"done wiht "<<i<<" iteration"<<endl;
         cout<<"dominators"<<endl;
         __gnu_cxx::hash_map<unsigned int, Tuple*, GnuHash>::iterator itt;	        
         for (unsigned int s=0;s<Di.size();s++)
         {    cout<<Di[s].sortid<<" ";
              itt=hatDi.find(Di[s].id);
              Tuple* t=(*itt).second;
              cout<<(*t)<<" "<<t->qitj<<endl;
         }
         cout<<"largest_rhoi "<<largest_rhoi<<endl;
         cout<<"terminator "<<terminator<<endl;
         cout<<endl;
         */

         if ( ( (topk.size()==(unsigned int)k) && (largest_rhoi>= terminator) ) || (i>=db->N-1) )
         {
             this->n=i+1;
             this->L_rhoi=largest_rhoi;
             break;
         }
         i++;
   }
   finalizeFromSnapshot();
   gettimeofday(&end,NULL);
   t2 = times(&ed);
   usertime=((double)(ed.tms_utime-st.tms_utime))/(HZ);
   systemtime=((double)(ed.tms_stime-st.tms_stime))/(HZ);
   realtime=((double)(t2-t1))/(HZ);
   db->timeval_subtract(&tdiff, &end, &start);
   elapsed=tdiff.tv_sec+tdiff.tv_usec/1000000.0;
   getMemsize();
}

void UtopkCor::printStatistics()
{	
    cout<<endl<<"scan depth: "<<this->n<<" "<<"utopk prob: "<<this->L_rhoi<<endl<<endl;
    printf("user time=%f \n", usertime);
    printf("system time=%f \n", systemtime);
    printf("real time=%f \n", realtime);
    printf("clock time=%f \n", elapsed); 
    cout<<"max mem usage(in bytes)= "<<memusage<<endl;
    cout<<"max mem usage(in KB)= "<<(double)memusage/1024.0<<endl;
    cout<<"max mem usage(in MB)= "<<(double)memusage/(1024.0*1024.0)<<endl;
}


void UtopkCor::printQueue(priority_queue<Tuple*, vector<Tuple*>, PrioritizeTask> res)
{
     cout<<"scan depth: "<<this->n<<endl;
      while ( !res.empty() ) 
      { 
            Tuple* tp=res.top();
            cout << (*tp) << " "<<tp->qitj<<endl; 
            res.pop(); 
      } cout << endl;     
}

void UtopkCor::finalizeFromSnapshot()
{
     map<Tuple*, Tuple*, PrioritizeTask>::iterator it_topk;	 
     for (unsigned int i=0;i<update_snapshot_topk.size();i++)
     {
             Tuple* tp=update_snapshot_topk[i];
             snapshot_topk.insert(pair<Tuple*, Tuple*>(tp, tp));
             it_topk=snapshot_topk.end();
             it_topk--;             
             snapshot_topk.erase(it_topk);
             if (tp->sortid==largest_rhoi_index)
                break;
     }
}


void UtopkCor::printVector(priority_queue<Tuple*, vector<Tuple*>, PrioritizeTask> res)
{
      vector<Tuple> topkres;
      while ( !res.empty() ) 
      { 
            Tuple* tp=res.top();
            res.pop(); 
            topkres.push_back(*tp);
            
      }
      sort(topkres.begin(),topkres.end(),compare_tuple_score1); 
      for (int i=0;i<k;i++)
      {
          Tuple tp=topkres[i];
          cout << tp << " "<<tp.qitj<<" "<<tp.confidence/tp.qitj<<endl;
      }
}

void UtopkCor::test_priority_queue(priority_queue<Tuple*, vector<Tuple*>, PrioritizeTask> res)
{
     Tuple* tp1=new Tuple(1, 1, 10, 0.5);
     tp1->qitj=0.5;
          
     Tuple* tp2=new Tuple(2, 2, 20, 0.5);
     tp2->qitj=0.25;
     
     Tuple* tp3=new Tuple(3, 3, 30, 0.5);
     tp3->qitj=0.05;
     
     res.push(tp1);
     res.push(tp2);
     res.push(tp3);
     
     Tuple* tp=res.top();
     cout<<(*tp)<<" "<<tp->qitj<<endl;
   
     res.pop();
     
     tp=res.top();
     cout<<(*tp)<<" "<<tp->qitj<<endl;

     res.pop();
     tp=res.top();
     cout<<(*tp)<<" "<<tp->qitj<<endl;
     
     cout<<endl;

}

void UtopkCor::printVector(map<Tuple*, Tuple*, PrioritizeTask> res)
{
      vector<Tuple> topkres;
      map<Tuple*, Tuple*, PrioritizeTask>::iterator it;
      it = res.begin();
      while (!res.empty())
      {
            Tuple* tp=(*it).second;
            topkres.push_back(*tp);
            res.erase(it);
            it=res.begin();
      }	        
      sort(topkres.begin(),topkres.end(),compare_tuple_score1); 
      for (int i=0;i<k;i++)
      {
          Tuple tp=topkres[i];
          //cout << tp << " "<<tp.qitj<<" "<<tp.confidence/tp.qitj<<endl;
          cout << tp <<endl;
      }
}

void UtopkCor::test_map(map<Tuple*, Tuple*, PrioritizeTask> res)
{
     Tuple* tp1=new Tuple(1, 1, 10, 0.5);
     tp1->qitj=0.5;
          
     Tuple* tp2=new Tuple(2, 2, 20, 0.5);
     tp2->qitj=0.05;
     
     Tuple* tp3=new Tuple(3, 3, 30, 0.5);
     tp3->qitj=0.5;
     
     res.insert(pair<Tuple*, Tuple*>(tp1, tp1));
     res.insert(pair<Tuple*, Tuple*>(tp2, tp2));
     res.insert(pair<Tuple*, Tuple*>(tp3, tp3));
     
     map<Tuple*, Tuple*, PrioritizeTask>::iterator it;
     
     it = res.end();
 
     it--;
     Tuple* tp=(*it).second;
     cout<<(*tp)<<" "<<tp->qitj<<endl;
 
     it--;
     tp=(*it).second;
     cout<<(*tp)<<" "<<tp->qitj<<endl;
     
     it--;
     tp=(*it).second;
     cout<<(*tp)<<" "<<tp->qitj<<endl<<endl;
     
     
     it=res.find(tp1);
     tp=(*it).second;
     cout<<(*tp)<<" "<<tp->qitj<<endl;
     
     res.erase(it);
     
     cout<<(*tp1)<<" "<<tp1->qitj<<endl;
     
     /*
     it=res.begin();
     tp=(*it).second;
     cout<<(*tp)<<" "<<tp->qitj<<endl;

     res.erase(it);
     it=res.begin();
     tp=(*it).second;
     cout<<(*tp)<<" "<<tp->qitj<<endl;

     cout<<endl;*/
     //tp1->qitj=0.01;
}

void UtopkCor::printResults()
{
     printVector(snapshot_topk);
}

void UtopkCor::getMemsize()
{
     //it is dominated by the hatDi size times tuple size (with qitj) at termination time
     memusage=hatDi.size()*Tuple::size(true);
     memusage+=k*Tuple::size(true);
}


