//coded by Feifei Li, contact: lifeifei@cs.bu.edu

#include "utopkindep.h"

bool compare_tuple_score1(Tuple t1, Tuple t2)
{
    return t1.sortid<t2.sortid;       
} 


UtopkIndep::UtopkIndep(int k)
{
   this->k=k;    
   this->db=NULL;
   elapsed=0.0;                       
}

UtopkIndep::~UtopkIndep()
{
   if (db!=NULL)                      
      delete db;                         
}

bool UtopkIndep::initDB(unsigned int N, char* input_file)
{
   db= new Udb(N, input_file);
   if (!(db->fin))
      return false;
   db->readRecords();
   //db->sortByScore();
   return true;     
}

void UtopkIndep::findTopk()
{
   unsigned int i=0;
   long double terminator=1.0;
   long double largest_rhoi=0.0;
   largest_rhoi_index=0;
   long double rhoi=1.0;
   
   t1 = times(&st);
   gettimeofday (&start,NULL);
   while (true)
   {
         Tuple tp=db->tuples[i];
         if (tp.confidence>=0.5)     
                 terminator=terminator*tp.confidence;
         else
                 terminator=terminator*(1-tp.confidence);                 
         if (i<(unsigned int)k)
         {
                 topk.push(tp);
                 rhoi=rhoi*tp.confidence;
                 i++;
                 if ( i==(unsigned int)k )
                 {
                        snapshot_topk=topk;
                        largest_rhoi=rhoi;
                        if (largest_rhoi>=terminator)
                        {
                              this->L_rhoi=largest_rhoi;                     
                              this->n=i;
                              break;                         
                        }
                 }
                 continue;
         }
         else
         {
             topk.push(tp);
             Tuple poped=topk.top();
             topk.pop();
             rhoi=(rhoi/poped.confidence)*tp.confidence*(1-poped.confidence);
             //if (update_snapshot_topk.size()<(unsigned int)k)
             if (poped.sortid!=tp.sortid)
                update_snapshot_topk.push_back(tp.sortid);
             if (rhoi>=largest_rhoi)
             {
                   if (update_snapshot_topk.size()>=(unsigned int)k)
                   {
                        update_snapshot_topk.clear();
                        snapshot_topk=topk;     
                   }      
                   largest_rhoi=rhoi;
                   largest_rhoi_index=tp.sortid;
                   this->L_rhoi=largest_rhoi;
             }
             i++;
             if (largest_rhoi>=terminator)
             {
                this->n=i;
                this->L_rhoi=largest_rhoi;
                break;                         
             }
             if (i>=db->N)
             {
                this->n=db->N;
                this->L_rhoi=largest_rhoi;
                break;
             }
         }
   }
   finalizeFromSnapshot();
   gettimeofday (&end,NULL);
   t2 = times(&ed);
   usertime=((double)(ed.tms_utime-st.tms_utime))/(HZ);
   systemtime=((double)(ed.tms_stime-st.tms_stime))/(HZ);
   realtime=((double)(t2-t1))/(HZ);
   db->timeval_subtract(&tdiff, &end, &start);
   elapsed=tdiff.tv_sec+tdiff.tv_usec/1000000.0;
   getMemsize();
}

void UtopkIndep::printStatistics()
{	
    //this->L_rhoi.adjustAB();
    cout<<endl<<"scan depth: "<<this->n<<" "<<"utopk prob: "<<this->L_rhoi<<endl<<endl;
    printf("user time=%f \n", usertime);
    printf("system time=%f \n", systemtime);
    printf("real time=%f \n", realtime);
    printf("clock time=%f \n", elapsed); 
    cout<<"max mem usage(in bytes)= "<<memusage<<endl;
    cout<<"max mem usage(in KB)= "<<(double)memusage/1024.0<<endl;
    cout<<"max mem usage(in MB)= "<<(double)memusage/(1024.0*1024.0)<<endl;
}

void UtopkIndep::printQueue(priority_queue<Tuple, vector<Tuple>, PrioritizeTask> res)
{
     cout<<"scan depth: "<<this->n<<endl;
      while ( !res.empty() ) 
      { 
            Tuple tp=res.top();
            cout << tp << endl; 
            res.pop(); 
      } cout << endl;     
}

void UtopkIndep::printVector(priority_queue<Tuple, vector<Tuple>, PrioritizeTask> res)
{
      vector<Tuple> topkres;
      while ( !res.empty() ) 
      { 
            Tuple tp=res.top();
            topkres.push_back(tp);
            res.pop(); 
      }
      sort(topkres.begin(),topkres.end(),compare_tuple_score1); 
      for (int i=0;i<k;i++)
      {
          Tuple tp=topkres[i];
          cout << tp <<endl;
      }
}

void UtopkIndep::finalizeFromSnapshot()
{
     for (unsigned int i=0;i<update_snapshot_topk.size();i++)
     {
             Tuple tp=db->tuples[update_snapshot_topk[i]];
             if (tp.sortid>largest_rhoi_index)
                break;
             //cout<<tp<<endl;
             snapshot_topk.push(tp);
             snapshot_topk.pop();
     }     
}

void UtopkIndep::printResults()
{
      printVector(snapshot_topk);
      //cout<<"topk"<<endl;
      //printVector(topk);
}

void UtopkIndep::getMemsize()
{
     //for utopkindep, the largest mem consumption is bounded by number of tuples stored in the priority queue, so ..k*tpsize     
     memusage=k*Tuple::size(false);
     memusage+=k*8; //priority_queue
     memusage+=k*4;
}

void UtopkIndep::stateProb(vector<unsigned int> tps, vector<Tuple> tuples)
{
     double prob=1.0;
     unsigned int i=0,j=0;
     while (i<tps.size())
     {
           if (tuples[j].sortid<tps[i])
              prob=prob*(1-tuples[j].confidence);
           else if (tuples[j].sortid==tps[i])
           {
              prob=prob*tuples[j].confidence;
              i++;
           }
           j++;               
     }
     cout<<"prob is "<<prob<<endl;
}
