//coded by Feifei Li, contact: lifeifei@cs.bu.edu


#include "state.h"
#include <stdio.h>
#include <stdlib.h>
#include <iostream>
#include<assert.h>

ostream & operator<<( ostream &os, State &s ) 
{ 
        os << s.length <<" " << s.prob; 
        return ( os ); 
}

State::State()
{  
   this->length=0;
   this->prob=0.0;
   this->last_seen_sortid=0;
   this->xtuple=false;
   this->last_sortid=0;
   this->init=true;
   this->prevstate=NULL;
}

State::State(bool xtuple)
{  
   this->length=0;
   this->prob=0.0;
   this->last_seen_sortid=0;
   this->xtuple=xtuple;
   this->last_sortid=0;
   this->init=true;
   this->prevstate=NULL;
}


State::State(unsigned int length, double prob, unsigned int last_seen_sortid, bool xtuple)
{  
   this->length=length;
   this->prob=prob;
   this->last_seen_sortid=last_seen_sortid;
   this->xtuple=xtuple;
   this->last_sortid=0;
   this->init=true;
   this->prevstate=NULL;
}

State::State(State& s)
{
   this->init=s.init;
   this->length=s.length;
   this->prob=s.prob;
   this->last_seen_sortid=s.last_seen_sortid;
   this->last_sortid=s.last_sortid;
   this->xtuple=s.xtuple;
   this->prevstate=s.prevstate;
   this->xtp_ids=s.xtp_ids;
}

State::~State()
{
    //tp_sortid.clear();
    xtp_ids.clear();
}


void State::set_prevstate(State* s)
{
     this->prevstate=s;     
}

int State::findXids(unsigned int xid)
{
    int cursize=this->xtp_ids.size();
    if (cursize==0)
       return -1;
    int low=0;
    int high=cursize-1;
    int mid=0;
    while(low<=high)
    {
       mid=(low+high)/2;
       if (xtp_ids[mid].id>xid)
          high=mid-1;
       else if (xtp_ids[mid].id<xid)
            low=mid+1;
       else
           return mid;
    }
    return -1;          
}

void State::tps_in_state(vector<Tuple> tuples)
{
     vector<int> tps;
     State* cur=this;
     State* tmp;
     while(cur!=NULL)
     {
           if (xtuple==false)
           {
               if (cur->last_seen_sortid==cur->last_sortid && cur->length!=0)
               {
                     tps.push_back(cur->last_sortid);
               }
           }    
           else
           {             
                  if (cur->last_seen_sortid==cur->last_sortid && cur->length!=0)
                  {                                          
                       int index=findXids(tuples[cur->last_sortid].id);
                       if (index==-1)
                            tps.push_back(cur->last_sortid);
                       else
                       {
                           if (xtp_ids[index].sortid==cur->last_sortid)    
                               tps.push_back(cur->last_sortid);
                       }     
                  }
           }
           tmp=cur;
           cur=cur->prevstate;
           if (tmp!=this)
             delete tmp;
     }
     for (int i=tps.size()-1; i>=0;i--)
         cout<<tuples[tps[i]]<<endl;          
}

void State::update_state_notp(Tuple tp, bool tp_couldbe_instate, double eff_confidence)
{
   if (this->init==true)
   {
      this->last_seen_sortid=tp.sortid;
      this->prob=eff_confidence;
      this->init=false;
   }
   else
   {
      this->last_seen_sortid=tp.sortid;
      if (tp_couldbe_instate)
            this->prob=this->prob*eff_confidence;    
   }
}

void State::update_state(Tuple tp, bool isxtuple, int index)
{
     this->last_seen_sortid=tp.sortid;
     this->last_sortid=tp.sortid;
     if (xtuple==false)
        this->length++;
     else
        this->length++;   
     if (isxtuple)
     {
         if (index!=-1)
         {
            xtp_ids[index].sortid=tp.sortid;
         }
         else
         {
            Tpids xtp;
            xtp.id=tp.id;
            xtp.sortid=tp.sortid; 
            xtp_ids.push_back(xtp);
         }
     }   
     //note for xtuple dbs, remeber divide old(tp's preceeding) alternative's confidence before calling update_state!!!
     if (this->init==true)
     {
        this->prob=tp.confidence;
        this->init=false;
     }
     else
        this->prob=this->prob*tp.confidence;
}


unsigned long State::getSize()
{
      /*state stores tp's id (or pointer) which is 4 bytes. that is the dominator factor. plus the 8 bytes of prob, and 4 bytes of last_seen_id
      if (xtuple)
            return (8+4+this->length*4);
      else
            return (8+4+this->length*4);         */
       //8 bytes of prob, 4 bytes of last_seen_id, 4 bytes of last_sortid, 4 bytes of pointer to prevstate, 4 bytes of length
      if (!xtuple)
            return (8+4+4+4+4);
      else
            return (8+4+4+4+4+this->xtp_ids.size()*8);     
}

void State::printXtpids()
{
     cout<<"printing..."<<endl;
     for (unsigned int i=0;i<xtp_ids.size();i++)
         cout<<xtp_ids[i].id<<" "<<xtp_ids[i].sortid<<endl;
     cout<<"done..."<<endl;
}
