//coded by Feifei Li, contact: lifeifei@cs.bu.edu

#include "udb.h"

bool compare_tuple_score(Tuple t1, Tuple t2)
{
    return t1.score>=t2.score;       
} 


Udb::Udb(unsigned int N, char* file_input)
{
   this->N=N;
   fin.open(file_input);                   
   if (!fin)
    	std::cerr << "Cannot open file " << file_input << std::endl;        
}

Udb::~Udb()
{
   if (fin.is_open())
      fin.close(); 
}

bool Udb::readRecords()
{ 
    if (!fin)
       return false;
    unsigned int counter=0;
    unsigned int score;
    unsigned int id;
    double confidence;
    while(fin)
    {
              fin >> id >> score >> confidence;
              if (confidence<0.0)
                 confidence=(1+confidence)*0.01;
              else if (confidence>1.0)
                 confidence=0.95;
              if (score<=0)
                 score=1;
              if (!fin.good()) continue; //skip newlines, etc
              Tuple tp(counter, id, score, confidence);
              tuples.push_back(tp);
              counter++;
              if (counter>=N)
                 break;
    }           
    return true;  
}

void Udb::sortByScore()
{
     sort(tuples.begin(),tuples.end(),compare_tuple_score);     
}

void Udb::printDB()
{
     for (unsigned int i=0;i<N;i++)
         //tuples[i].show();
         cout<<tuples[i]<<endl;     
}

unsigned int Udb::buildXtuples(char* file_input, int xdegree, unsigned int numxtuple)
{
   std::ifstream xfin;
   xfin.open(file_input);                   
   if (!xfin)
   {
    	std::cerr << "Cannot open file " << file_input << std::endl;   
    	this->numxtuple=0;
    	this->xdegree=0;
    	return 0;
   }
   xtuples.resize(this->N);	
  	unsigned int counter=0;
  	int i=0,j=0;
  	unsigned int xtpid[xdegree];
  	
    while(xfin)
    {
              for (i=0;i<xdegree;i++)
              {
                  xfin >> xtpid[i];
              }
              if (!xfin.good()) continue; //skip newlines, etc
              for (i=0;i<xdegree;i++)
              {
                  for (j=0;j<xdegree;j++)
                  {
                      if (i!=j)
                         xtuples[xtpid[i]-1].push_back(xtpid[j]-1);
                  }
              }
              counter++;
    }
    if (counter!=numxtuple)
       cout<<"wrong... number of xtuples no equal to the specified value in the file"<<endl;
    this->numxtuple=numxtuple;
    this->xdegree=xdegree;
    return counter;
}

void Udb::printXtuple(unsigned int id)
{
     //only call this function after buildXtuple
     cout<<id<<" its xtuples are: ";
     if (xtuples[id].size()!=0)
     {
        for (int i=0;i<xdegree-1;i++)
             cout<< tuples[xtuples[id][i]]<<" ";
        cout<<endl;
     }
}

bool Udb::has_preceeding_alternative(unsigned int tp_sortid, bool& isxtuple)
{
     bool have=true;
     unsigned int x_sortid;
     if (xtuples[tp_sortid].size()==0 )
     {
        have=false;
        isxtuple=false;
     }
     else
     {
         //as xid list is sorted, check the first one is good enough;
         x_sortid=xtuples[tp_sortid][0];
         if (x_sortid<tp_sortid)
            have=true;
         else
            have=false;
         isxtuple=true;
     }
     return have;
}

int Udb::timeval_subtract(struct timeval* result, struct timeval* x, struct timeval* y)
{
	  /* Perform the carry for the later subtraction by updating y. */
	  if (x->tv_usec < y->tv_usec) {
	    int nsec = (y->tv_usec - x->tv_usec) / 1000000 + 1;
	    y->tv_usec -= 1000000 * nsec;
	    y->tv_sec += nsec;
	  }
	  if (x->tv_usec - y->tv_usec > 1000000) {
	    int nsec = (y->tv_usec - x->tv_usec) / 1000000;
	    y->tv_usec += 1000000 * nsec;
	    y->tv_sec -= nsec;
	  }
	
	  /* Compute the time remaining to wait.
	     tv_usec  is certainly positive. */
	  result->tv_sec = x->tv_sec - y->tv_sec;
	  result->tv_usec = x->tv_usec - y->tv_usec;
	
	  /* Return 1 if result is negative. */
	  return x->tv_sec < y->tv_sec;
}


