//coded by Feifei Li, contact: lifeifei@cs.bu.edu

#include "ukrankcor.h"

UkrankCor::UkrankCor(int k)
{
   this->k=k;    
   this->db=NULL;
   elapsed=0.0;                       
}

UkrankCor::~UkrankCor()
{
   if (db!=NULL)
      delete db;                                                   
}

bool UkrankCor::initDB(unsigned int N, char* input_file, char* xtuple_file, int xdegree, unsigned int numxtuple)
{
   db= new Udb(N, input_file);
   if (!(db->fin))
      return false;
   db->readRecords();
   db->buildXtuples(xtuple_file, xdegree, numxtuple);
   //db->sortByScore(); not necessary if the input db has already sorted, which is the case for our generator.
   return true;     
}


void UkrankCor::findTopk()
{
   unsigned int i=0;
   unsigned int depth=0;
   unsigned int j=0;
   unsigned int index;
   bool isxtuple=false;
   
   memusage=0;
   maxnum_states=0;
   
   i=1;
   report=0;
   topk.resize(k+1);
   krank.resize(k+1);
   Pt.resize(k+1);
   ubounds.resize(k+1);
   reported.resize(k+1);
   states.resize(k+1);
   for (j=0;j<(unsigned int)k+1;j++)
   {
       topk[j]=0;
       krank[j]=0.0;
       ubounds[j]=1;
       Pt[j]=0.0;
       reported[j]=false;
   }
   
   //State* s0= new State(0, 1.0, 0, true);
   //states[0].push_back(s0);
   //states.insert(pair<unsigned int, State*>(0, s0));
   
   t1 = times(&st);
   gettimeofday (&start,NULL);
   Tuple tp=db->tuples[0];
   db->has_preceeding_alternative(tp.sortid,isxtuple);
   State* s0=new State(0,1.0,0,true);
   State* s1=new State(0,1.0,0,true);
   
   s0->update_state_notp(tp,true,1-tp.confidence);
   s1->update_state(tp,isxtuple,-1);
   states[0].push_back(s0);
   states[1].push_back(s1);
   
   krank[1]=s1->prob;
   topk[1]=tp.sortid;
   if (s1->prob>s0->prob)
   {
        report++;
        reported[1]=true;
   }
   depth++;

   while (depth<db->N && report<k)
   { 
          Tuple tp=db->tuples[depth];
          expandStates(tp);
          
          index=(unsigned)k<(depth+1)?k:(depth+1);
          for (i=1;i<=index;i++)
          {
               if (reported[i]==false)
               {
                  if (Pt[i]>krank[i])
                  {
                          krank[i]=Pt[i];
                          topk[i]=tp.sortid;
                  }
                  if (Pt[i]>ubounds[i])
                  {
                          reported[i]=true;
                          report++;
                  }
               }
          }
          pruneStates();
          depth+=1;
   }
   fflush(stdout);
   this->n=depth;
   memusage+=this->k*Tuple::size(false);
   gettimeofday (&end,NULL);
   t2 = times(&ed);
   usertime=((double)(ed.tms_utime-st.tms_utime))/(HZ);
   systemtime=((double)(ed.tms_stime-st.tms_stime))/(HZ);
   realtime=((double)(t2-t1))/(HZ);
   db->timeval_subtract(&tdiff, &end, &start);
   elapsed=tdiff.tv_sec+tdiff.tv_usec/1000000.0;
}

void UkrankCor::printStatistics()
{	
    cout<<endl<<"scan depth: "<<this->n<<endl<<endl;
    printf("user time=%f \n", usertime);
    printf("system time=%f \n", systemtime);
    printf("real time=%f \n", realtime);
    printf("clock time=%f \n", elapsed); 
    cout<<"max mem usage(in bytes)= "<<memusage<<endl;
    cout<<"max mem usage(in KB)= "<<(double)memusage/1024.0<<endl;
    cout<<"max mem usage(in MB)= "<<(double)memusage/(1024.0*1024.0)<<endl;
    cout<<"max number of states= "<<maxnum_states<<endl;
}

void UkrankCor::printStateSize()
{
     for (int i=0;i<=k;i++)
         cout<<states[i].size()<<endl;
}

void UkrankCor::printStates(unsigned int i)
{
    for (unsigned int j=0;j<states[i].size();j++)
    {
        cout<<(*states[i][j])<<endl;
        states[i][j]->tps_in_state(db->tuples);
        cout<<endl;
    }      
    cout<<endl;
}

void UkrankCor::expandStates(Tuple tp)
{
     unsigned int i=0;
     unsigned int j=0;
     unsigned long numstates=0;
     unsigned long curmem=0;
     bool isxtuple=false;
     int index=-1;
   
     for (i=0;i<=(unsigned int)k;i++)
     {
         if (i<(unsigned int)k)
         {
             Pt[i+1]=0.0;
             ubounds[i+1]=0.0;       
         }
         if (states[i].size()!=0)
         {
            for (j=0;j<states[i].size();j++)
            {
                if (states[i][j]->last_seen_sortid==tp.sortid-1)
                {
                         if (db->has_preceeding_alternative(tp.sortid,isxtuple)==false)
                         {
                               State* s= new State((*states[i][j]));
                              
                               states[i][j]->update_state_notp(tp, true,1-tp.confidence);
                               
                               s->update_state(tp,isxtuple,-1);
                               s->set_prevstate(states[i][j]);
                               if (i<(unsigned int)k)
                               {
                                   states[i+1].push_back(s);
                                   Pt[i+1]=Pt[i+1]+s->prob;
                                   ubounds[i+1]=ubounds[i+1]+states[i][j]->prob;
                               }
                         }
                         else
                         {
                              index = states[i][j]->findXids(tp.id);
                              if (index!=-1)
                              {
                                 states[i][j]->update_state_notp(tp,false,1-tp.confidence);
                                 if (i<(unsigned int)k)
                                 {
                                    ubounds[i+1]=ubounds[i+1]+states[i][j]->prob;
                                 }
                              }
                              else
                              {
                                 double sum=0.0;
                                 for (unsigned int e=0;e<db->xtuples[tp.sortid].size();e++)
                                 {
                                     unsigned int xid= db->xtuples[tp.sortid][e];
                                     if (xid>tp.sortid)
                                        break;
                                     sum+=db->tuples[xid].confidence;
                                 }
                                 State* s= new State(*(states[i][j]));

                                 states[i][j]->prob=states[i][j]->prob/(1-sum);
                                 states[i][j]->update_state_notp(tp, true, 1-sum-tp.confidence);
                 
                                 s->prob=s->prob/(1-sum);
                                 s->update_state(tp,isxtuple,index);
                                 s->set_prevstate(states[i][j]);
                                 if (i<(unsigned int)k)
                                 {
                                    states[i+1].push_back(s);
                                    Pt[i+1]=Pt[i+1]+s->prob;
                                    ubounds[i+1]=ubounds[i+1]+states[i][j]->prob;
                                 }
                             }       
                         }
                 }
                 else
                 {
                         if (i<(unsigned int)k)
                            ubounds[i+1]+=states[i][j]->prob;
                 }
                 curmem+=states[i][j]->getSize();
            }
            numstates+=states[i].size();
        }
     }
     if (numstates>maxnum_states)
        maxnum_states=numstates;
     if (curmem>memusage)
        memusage=curmem;
}

bool UkrankCor::checkTerminate(vector<double> krank, vector<double>Rij)
{
     int i=0;
     bool stop=true;
     double maxrnl=0.0;
     for (i=1;i<=this->k;i++)
     {
         if (maxrnl<Rij[i-1])
              maxrnl=Rij[i-1];
         if (krank[i]<maxrnl)
         {
              stop=false;
              break;                      
          }  
     }
     return stop;
}

void UkrankCor::printTopk()
{
     int i=0;
     for (i=1;i<=k;i++)
     {
         Tuple tp=db->tuples[topk[i]];
         cout<<tp<<" "<<krank[i]<<endl;
     }    
}

void UkrankCor::test_hashmap(__gnu_cxx::hash_map<unsigned int, unsigned int, GnuHash> mp)
{
     mp.insert(pair<unsigned int, unsigned int>(10, 2));
     mp.insert(pair<unsigned int, unsigned int>(3, 5));
     mp.insert(pair<unsigned int, unsigned int>(2, 4));
     
     __gnu_cxx::hash_map<unsigned int, unsigned int, GnuHash>::iterator it;	      
     it=mp.begin();
     cout<<(*it).first<<" "<<(*it).second<<endl; 
     //(*it).second=7;
     it++;
          cout<<(*it).first<<" "<<(*it).second<<endl;
          it++;
               cout<<(*it).first<<" "<<(*it).second<<endl;
     
     it=mp.begin();
     cout<<(*it).first<<" "<<(*it).second<<endl; 
     
     vector<double> p1;
     vector<double> p2;
     p2.resize(2);
     
     p1.push_back(0.4);
     p1.push_back(0.3);
     cout<<"rij results: "<<endl;
     //cout<<p2[0]<<endl;
     //cout<<p2[1]<<endl;
     //cout<<p2[2]<<endl;
}

void UkrankCor::printPtiUbounds()
{
     for (int i=1;i<=k;i++)     
         cout<<Pt[i]<<" "<<ubounds[i]<<endl;
}

void UkrankCor::pruneStates()
{

    int highest=0;    
    int i;      
    for (i=k; i>=0; i--)
    {   
        if (reported[i]==false)
           break;
    }
    highest=i;
    for (int i=k;i>highest;i--)
    {
        for (unsigned int j=0;j<states[i].size();j++)
        {
            if (states[i][j]!=NULL)
                        delete states[i][j];    
        }
        states[i].clear();
    }
}
