//coded by Feifei Li, contact: lifeifei@cs.bu.edu

#include "ukrankindep.h"

UkrankIndep::UkrankIndep(int k)
{
   this->k=k;    
   this->db=NULL;
   elapsed=0.0;                       
}

UkrankIndep::~UkrankIndep()
{
   if (db!=NULL)
      delete db;                                                   
}

bool UkrankIndep::initDB(unsigned int N, char* input_file)
{
   db= new Udb(N, input_file);
   if (!(db->fin))
      return false;
   db->readRecords();
   //db->sortByScore();
   return true;     
}

void UkrankIndep::findTopk()
{
   unsigned int i=0;
   unsigned int j=0;
   long s=0;
   //long r=0;
   bool stop=false;
   double allnonprev=1.0;

   M.resize(k);
   topk.resize(k);
   rankprob.resize(k);
   i=0;
   t1 = times(&st);
   gettimeofday (&start,NULL);
   while (true)
   {
          Tuple tp=db->tuples[i];
          for (j=0;j<(unsigned int)k;j++)
          {
              if (j==0)
              {
                       double prob=1.0;
                       //for (s=0;s<(long)i;s++)   
                           //prob=prob*(1-db->tuples[s].confidence);
                       prob=prob*allnonprev;
                       prob=prob*tp.confidence;
                       M[j].push_back(prob);
              }
              else
              {   
                       double prob=0.0;
                       double nonprev=1.0;
                       for (s=i-1;s>=0;s--)
                       {
                           double My=M[j-1][s];
                           double product=1.0;
                           //for (r=i-1;r>s;r--)
                                //product*=(1-db->tuples[r].confidence);                         
                           product=product*nonprev;     
                           nonprev*=(1-db->tuples[s].confidence);     
                           prob+=My*product;
                       }
                       prob=prob*tp.confidence;
                       M[j].push_back(prob);
              }
          }
          //cout<<"done with iteration "<<i<<endl;
          //printM();
          allnonprev*=1-db->tuples[i].confidence;
          if (i<(unsigned int)k-1)
             stop=false;
          else
              stop=checkTerminate();
          if (stop || i>=db->N-1)
          {
                   this->n=i++;
                   break;
          }
          i++;
   }
   gettimeofday (&end,NULL);
   t2 = times(&ed);
   usertime=((double)(ed.tms_utime-st.tms_utime))/(HZ);
   systemtime=((double)(ed.tms_stime-st.tms_stime))/(HZ);
   realtime=((double)(t2-t1))/(HZ);
   db->timeval_subtract(&tdiff, &end, &start);
   elapsed=tdiff.tv_sec+tdiff.tv_usec/1000000.0;
   getMemsize();
}

void UkrankIndep::printStatistics()
{	
    cout<<endl<<"scan depth: "<<this->n<<endl;
    printf("user time=%f \n", usertime);
    printf("system time=%f \n", systemtime);
    printf("real time=%f \n", realtime);
    printf("clock time=%f \n", elapsed); 
    cout<<"max mem usage(in bytes)= "<<memusage<<endl;
    cout<<"max mem usage(in KB)= "<<(double)memusage/1024.0<<endl;
    cout<<"max mem usage(in MB)= "<<(double)memusage/(1024.0*1024.0)<<endl;
}

bool UkrankIndep::checkTerminate()
{
     int i=0;
     unsigned int j=0;
     bool stop=true;
     for (i=0;i<k;i++)
     {
         unsigned int size = M[i].size();
         double max=0.0;
         double sum=0.0;
         unsigned int maxid=0;
         for (j=0;j<size;j++)
         {
               if (M[i][j]>max)
               {
                  max=M[i][j];
                  maxid=j;
               }
               sum+=M[i][j];
         }
         if ( max < (1-sum) )
            stop=false;
         topk[i]=maxid;
         rankprob[i]=max;
     }
     return stop;
}

void UkrankIndep::printM()
{
     int i=0;
     unsigned int j=0;
     for (i=0;i<k;i++)
     {
         unsigned int size = M[i].size();
         for (j=0;j<size;j++)
         {
              cout<<M[i][j]<<" ";
         }
         cout<<endl;
     }
}

void UkrankIndep::printTopk()
{
     
     int i=0;
     for (i=0;i<k;i++)
     {
         Tuple tp=db->tuples[topk[i]];
         cout<<tp<<" "<<rankprob[i]<<endl;
     }
     //cout<<endl<<"scan depth "<<this->n<<endl;
}

void UkrankIndep::getMemsize()
{
     //dominate by M (k * scandepth) + n (scan depth) tuples' confidence must be kept 
     
     memusage=this->n*this->k*8+n*8+this->k*Tuple::size(false);
}
