//coded by Feifei Li, contact: lifeifei@cs.bu.edu

#include "utopkcor.h"

bool compare_tuple_score1(Tuple t1, Tuple t2)
{
    return t1.score>=t2.score;       
} 


UtopkCor::UtopkCor(int k)
{
   this->k=k;    
   this->db=NULL;
   elapsed=0.0;                       
}

UtopkCor::~UtopkCor()
{
   if (db!=NULL)                      
      delete db;  
   while ( !pq_states.empty() ) 
   { 
            State* s=pq_states.top();
            delete s;
            pq_states.pop(); 
   } 
}

bool UtopkCor::initDB(unsigned int N, char* input_file, char* xtuple_file, int xdegree, unsigned int numxtuple)
{
   db= new Udb(N, input_file);
   if (!(db->fin))
      return false;
   db->readRecords();
   db->buildXtuples(xtuple_file, xdegree, numxtuple);
   //db->sortByScore(); not necessary if the input db has already sorted, which is the case for our generator.
   return true;     
}

void UtopkCor::findTopk(int showTopk)
{
   unsigned int depth=0;
   unsigned int xid=0;
   bool found=false;
   bool init=true;
   unsigned long queue_size=0;
   bool isxtuple=false;
   int index=-1;
   
   memusage=0;
   maxnum_states=0;
   t1 = times(&st);
   gettimeofday (&start,NULL);
   State* s0= new State(0, 1.0, 0, true);
   pq_states.push(s0);
   
   queue_size+=s0->getSize();
   memusage=queue_size;
   
   while (pq_states.empty()==false)
   {
         State* s=pq_states.top();
         pq_states.pop();
         queue_size-=s->getSize();
   
         Tuple tp;
         if (s->length==(unsigned int)k)
         {
            this->n=depth+1;
            found=true;
            res=new State(*s);
            //printTopk(s);
            break;
         }
         if (init)
         {
            tp=db->tuples[0];
         }
         else 
         {
              if (s->last_seen_sortid < (db->N-1) )
              {
                tp=db->tuples[s->last_seen_sortid+1];
              }                                 
              else
              {
                   //cout<<"no more extension to this state"<<endl;
                   delete s;
                   continue; 
              }
         }
         State* s1 = new State(*s);
         State* s2 = new State(*s);
         if (db->has_preceeding_alternative(tp.sortid,isxtuple)==false)
         {
              //it has no alternatives;
              //cout<<"no preceeding alters"<<tp.id<<" "<<tp.sortid<<endl;
              s1->update_state_notp(tp,true,1-tp.confidence);
              if (init==false)
                 s1->set_prevstate(s);
              s2->update_state(tp,isxtuple, -1);
              if (init==false)
                 s2->set_prevstate(s);
              
              pq_states.push(s1);
              pq_states.push(s2);
   
              queue_size+=s1->getSize();
              queue_size+=s2->getSize();
              if (showTopk==1)
                 queue_size+=s->getSize(); 
              if (showTopk==0) 
                 delete s;
              //cout<<"done1 updating s1 s2"<<endl;
         }
         else
         {
              //cout<<"searching "<<tp.id<<" "<<tp.sortid<<endl;
              //s2->printXtpids();
              index = s2->findXids(tp.id);
              //cout<<"index "<<index<<endl;
              if (index!=-1)
              {
                 //cout<<"have preceeding and in the state, so the current tp current appears in the state any more!! So we do nothing"<<endl;
                 //cout<<"found prev tuple in state "<<tp.id<<" "<<tp.sortid<<endl;
                 //s2->printXtpids();
                 s1->update_state_notp(tp, false, 1-tp.confidence);
                 if (init==false)
                     s1->set_prevstate(s);
                 pq_states.push(s1);
              
                 queue_size+=s1->getSize();
                 if (showTopk==1)
                    queue_size+=s->getSize();  
                 delete s2;
                 if (showTopk==0)
                    delete s;
             }
             else
             {
                 //cout<<"have preceeding but not in the state"<<endl;
                 //cout<<"found no prev tuple in state "<<tp.id<<" "<<tp.sortid<<endl;
                 //s2->printXtpids();
                 double sum=0.0;
                 for (unsigned int j=0;j<db->xtuples[tp.sortid].size();j++)
                 {
                     xid= db->xtuples[tp.sortid][j];
                     if (xid>tp.sortid)
                        break;
                     sum+=db->tuples[xid].confidence;
                 }
                 s1->prob=s1->prob/(1-sum);
                 s1->update_state_notp(tp,true, 1-sum-tp.confidence);
                 if (init==false)
                    s1->set_prevstate(s);
                 pq_states.push(s1);
                 queue_size+=s1->getSize();
                    
                 s2->prob=s2->prob/(1-sum);
                 s2->update_state(tp, isxtuple, index);
                 if (init==false)
                     s2->set_prevstate(s);
                 pq_states.push(s2);
                 queue_size+=s2->getSize();
                 if (showTopk==1)
                    queue_size+=s->getSize();  
                 if (showTopk==0)
                    delete s;
                 //cout<<"done updating s1 s2 "<<endl;
             }       
         }
         if (tp.sortid>=depth)
            depth=tp.sortid;
         if (queue_size>memusage)
            memusage=queue_size;
         if (pq_states.size()>maxnum_states)
            maxnum_states=pq_states.size();
         if (init)
            init=false;
   }
   memusage+=this->k*Tuple::size(false);
   gettimeofday (&end,NULL);
   t2 = times(&ed);
   usertime=((double)(ed.tms_utime-st.tms_utime))/(HZ);
   systemtime=((double)(ed.tms_stime-st.tms_stime))/(HZ);
   realtime=((double)(t2-t1))/(HZ);
   db->timeval_subtract(&tdiff, &end, &start);
   elapsed=tdiff.tv_sec+tdiff.tv_usec/1000000.0;

   if (found==false)
   {
      cout<<"We have exhausted the search space and no topk answer for k="<<k<<" has not been found.."<<endl;
      cout<<"scan depth is "<<n<<endl; 
   }
            
}

void UtopkCor::printStatistics()
{	
    cout<<endl<<"scan depth: "<<this->n<<" "<<"utopk prob: "<<res->prob<<endl<<endl;
    printf("user time=%f \n", usertime);
    printf("system time=%f \n", systemtime);
    printf("real time=%f \n", realtime);
    printf("clock time=%f \n", elapsed);
    cout<<"max mem usage(in bytes)= "<<memusage<<endl;
    cout<<"max mem usage(in KB)= "<<(double)memusage/1024.0<<endl;
    cout<<"max mem usage(in MB)= "<<(double)memusage/(1024.0*1024.0)<<endl;
    cout<<"max number of states= "<<maxnum_states<<endl;
}


void UtopkCor::printQueue(priority_queue<State*, vector<State*>, PrioritizeTask> res)
{
     cout<<"scan depth: "<<this->n<<endl;
      while ( !res.empty() ) 
      { 
            State* s=res.top();
            cout << (*s) <<endl; 
            res.pop(); 
      } cout << endl;     
}

void UtopkCor::printTopk(State* s)
{
     /*vector<Tuple> res;
     cout<<endl;
    __gnu_cxx::hash_map<unsigned int, unsigned int, GnuHash>::iterator it;	           
    it=s->xtp_sortid.begin();
    while(it!= s->xtp_sortid.end())
    {
          Tuple tp=db->tuples[(*it).second];
          //cout<<tp<<endl;
          res.push_back(tp);   
          it++; 
    }
    sort(res.begin(),res.end(),compare_tuple_score1);
    for (unsigned int i=0;i<res.size();i++)
        cout<<res[i]<<endl;*/
    //cout<<endl<<"scan depth is "<<this->n<<" and prob is "<<s->prob<<endl;
    s->tps_in_state(db->tuples);
}

void UtopkCor::test_priority_queue(priority_queue<State*, vector<State*>, PrioritizeTask> res)
{
     State* s1=new State();
     s1->prob=0.5;
     s1->length=2;
          
     State* s2=new State();
     s2->prob=0.5;
     s2->length=3;

     State* s3=new State();
     s3->prob=0.9;
     s3->length=4;
     
     res.push(s1);
     res.push(s2);
     res.push(s3);
     
     State* s=res.top();
     cout<<(*s)<<endl;
   
     res.pop();
     
     s=res.top();
     cout<<(*s)<<endl;

     res.pop();
     s=res.top();
     cout<<(*s)<<endl;
     
     cout<<endl;
}

void UtopkCor::testState()
{
   State s;
   
   Tpids x1;
   x1.id=0;
   x1.sortid=0; 
   s.xtp_ids.push_back(x1);
   
   Tpids x2;
   x2.id=1;
   x2.sortid=1; 
   s.xtp_ids.push_back(x2);
   
   Tpids x3;
   x3.id=2;
   x3.sortid=2; 
   
   s.xtp_ids.push_back(x3);
   
   int index=s.findXids(2);
   cout<<index<<endl;
   cout<<s.xtp_ids[index].id<<" "<<s.xtp_ids[index].sortid<<endl;
   s.xtp_ids[index].sortid=10;
   cout<<s.xtp_ids[index].id<<" "<<s.xtp_ids[index].sortid<<endl;
   index=s.findXids(10);
   cout<<index<<endl;
}
